"""
A package of python routines to handle webnucleo networks.
"""

__all__ = [
    "__title__",
    "__summary__",
    "__version__",
    "__author__",
    "__copyright__",
]

__title__ = "wnnet"
__summary__ = "Python project to handle webnucleo reaction networks"
__version__ = "2.6.1"
__author__ = "Clemson University"
__copyright__ = "Clemson University, 2022-2025"
